def byte_to_bit_list(byte: int):
    if not (0 <= byte <= 255):
        raise ValueError('Byte must be between 0 and 255!')

    bits = [(byte & (0x01 << i)) != 0 for i in range(8)]
    bits.reverse()
    return bits


def bit_list_to_byte(bits: list[bool]):
    if len(bits) != 8 or not all([isinstance(i, bool) for i in bits]):
        raise ValueError('Must provide exactly 8 bits (1 byte)!')

    byte = 0
    for bit in bits:
        byte = byte << 1 | (1 if bit else 0)
    return byte


def set_bit(byte: int, pos: int, val: bool):
    if not (0 <= byte <= 255):
        raise ValueError('Byte must be between 0 and 255!')
    if not (0 <= pos <= 7):
        raise ValueError('Position must be between 0 and 7!')
    if not isinstance(val, bool):
        raise ValueError('Bit must be a boolean!')

    mask = ~(1 << pos)
    return (byte & mask) | (val << pos)


def get_bit(byte: int, pos: int):
    if not (0 <= byte <= 255):
        raise ValueError('Byte must be between 0 and 255!')
    if not (0 <= pos <= 7):
        raise ValueError('Position must be between 0 and 7!')

    return (byte & (1 << pos)) != 0



from binary_tools import bit_encoder


def _string_len(data: bytes):
    pos = 0
    while len(data) > pos and data[pos] != 0x00:
        pos += 1
    return pos


class ByteEncoder:
    def __init__(self):
        self.__bit_buf = bit_encoder.BitEncoder()

    def put_string(self, string: str):
        binarised_str = string.encode('ascii') + b'\x00'
        self.__bit_buf.put_bytes(binarised_str)

    def put_int(self, number: int, length: int):
        binarised_int = int.to_bytes(number, length, 'big')
        self.__bit_buf.put_bytes(binarised_int)

    def put_bool(self, val: bool):
        self.__bit_buf.put_bits([val])

    def put_byte(self, byte: int, repeat: int = 1):
        for _ in range(repeat):
            converted_byte = int.to_bytes(byte, 1, 'big')
            self.__bit_buf.put_bytes(converted_byte)

    def get_data(self):
        return self.__bit_buf.finalise()


class ByteDecoder:
    def __init__(self, data: bytes):
        self.__buf = bit_encoder.BitDecoder(data)

    def has_more(self):
        return self.__buf.has_more()

    def consume_bytes(self, length: int):
        self.__buf.advance_bytes(length)

    def peek(self, length: int):
        return self.__buf.as_bytes()[:length]

    def get_string(self):
        length = _string_len(self.__buf.as_bytes())
        string = self.__buf.as_bytes()[0:length]
        self.__buf.advance_bytes(length + 1)

        return string.decode('ascii')

    def get_int(self, length: int):
        number = int.from_bytes(self.__buf.as_bytes()[0:length], 'big')
        self.__buf.advance_bytes(length)

        return number

    def get_bool(self):
        val = self.__buf.as_bit_list()[0]
        self.__buf.advance_bits(1)
        return val

    def get_byte(self):
        b = self.__buf.as_bytes()[0]
        self.__buf.advance_bytes(1)
        return b



import math

from binary_tools import binary_util


class BitEncoder:
    def __init__(self):
        self.__buf = [False, False, False, False]
        self.__data = b''

    def __encode(self):
        full_bytes = math.floor(len(self.__buf) / 8)

        for i in range(full_bytes):
            self.__data += binary_util.bit_list_to_byte(self.__buf[i*8:i*8+8]).to_bytes(1, 'big')
            self.__buf = self.__buf[i+8:]

    def put_bit(self, bit: bool):
        if not isinstance(bit, bool):
            raise ValueError('Bit must be a bool value!')

        self.__buf.append(bit)
        self.__encode()

    def put_bits(self, bits: list[bool]):
        if not isinstance(bits, list) or not all([isinstance(i, bool) for i in bits]):
            raise ValueError('Must provide list of bits (bool values)!')

        for bit in bits:
            self.put_bit(bit)

    def put_byte(self, byte: int):
        if not (0 <= byte <= 255):
            raise ValueError('Byte must be between 0 and 255!')

        self.put_bits(binary_util.byte_to_bit_list(byte))

    def put_bytes(self, bs: bytes):
        if not isinstance(bs, bytes):
            raise ValueError('Must provide a byte string!')

        for b in bs:
            self.put_byte(b)

    def finalise(self):
        self.__encode()

        remaining_bits = len(self.__buf)

        if remaining_bits == 0:
            return self.__data

        last_byte = 0x00
        for i in range(remaining_bits):
            last_byte = binary_util.set_bit(last_byte, 7 - i, self.__buf[i])
        self.__data += last_byte.to_bytes(1, 'big')

        head = remaining_bits << 4 | self.__data[0]

        return head.to_bytes(1, 'big') + self.__data[1:]


class BitDecoder:
    def __init__(self, data: bytes):
        if not isinstance(data, bytes):
            raise ValueError('Must provide a byte string!')

        self.__buf = []

        head = data[0]
        remaining_bits = head >> 4
        actual_head_bits = binary_util.byte_to_bit_list(head & 0x0F)[4:]
        self.__buf.extend(actual_head_bits)

        data = data[1:]

        for b in data:
            self.__buf.extend(binary_util.byte_to_bit_list(b))

        if remaining_bits > 0:
            self.__buf = self.__buf[:-(8-remaining_bits)]

    def advance_bits(self, count: int):
        self.__buf = self.__buf[count:]

    def advance_bytes(self, count: int):
        self.advance_bits(count * 8)

    def has_more(self):
        return len(self.__buf) > 0

    def as_bit_list(self):
        return self.__buf

    def as_bytes(self):
        full_bytes = math.floor(len(self.__buf) / 8)

        byte_string = b''
        for i in range(full_bytes):
            byte_string += binary_util.bit_list_to_byte(self.__buf[i*8:i*8 + 8]).to_bytes(1, 'big')

        return byte_string



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='binary_tools',
    version='0.1.0',
    author='Guilherme Borges',
    author_email='guilherme@guilhermeborges.net',
    description='Pyhton tools for binary data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sgtpepperpt/binary_tools',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9'
)
